@@ -1,6 +1,6 @@
 /* getkey.c -  Get a key from the database
- * Copyright (C) 1998, 1999, 2000, 2001, 2002,
- *               2003 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003,
+ *               2004 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -35,14 +35,13 @@
 #include "trustdb.h"
 #include "i18n.h"
 
-#define MAX_PK_CACHE_ENTRIES   200
-#define MAX_UID_CACHE_ENTRIES  200
+#define MAX_PK_CACHE_ENTRIES   PK_UID_CACHE_SIZE
+#define MAX_UID_CACHE_ENTRIES  PK_UID_CACHE_SIZE
 
 #if MAX_PK_CACHE_ENTRIES < 2
 #error We need the cache for key creation
 #endif
 
-
 struct getkey_ctx_s {
     int exact;
     KBNODE keyblock;
@@ -320,16 +319,21 @@ get_pubkey( PKT_public_key *pk, u32 *keyid )
     int rc = 0;
 
 #if MAX_PK_CACHE_ENTRIES
-    {	/* Try to get it from the cache */
+    if(pk)
+      {
+	/* Try to get it from the cache.  We don't do this when pk is
+	   NULL as it does not guarantee that the user IDs are
+	   cached. */
 	pk_cache_entry_t ce;
-	for( ce = pk_cache; ce; ce = ce->next ) {
-	    if( ce->keyid[0] == keyid[0] && ce->keyid[1] == keyid[1] ) {
-		if( pk )
-		    copy_public_key( pk, ce->pk );
+	for( ce = pk_cache; ce; ce = ce->next )
+	  {
+	    if( ce->keyid[0] == keyid[0] && ce->keyid[1] == keyid[1] )
+	      {
+		copy_public_key( pk, ce->pk );
 		return 0;
-	    }
-	}
-    }
+	      }
+	  }
+      }
 #endif
     /* more init stuff */
     if( !pk ) {
@@ -375,13 +379,15 @@ get_pubkey( PKT_public_key *pk, u32 *keyid )
 
 /* Get a public key and store it into the allocated pk.  This function
    differs from get_pubkey() in that it does not do a check of the key
-   to avoid recursion.  It should be used only in very certain cases.  */
+   to avoid recursion.  It should be used only in very certain cases.
+   It will only retrieve primary keys. */
 int
 get_pubkey_fast (PKT_public_key *pk, u32 *keyid)
 {
   int rc = 0;
   KEYDB_HANDLE hd;
   KBNODE keyblock;
+  u32 pkid[2];
   
   assert (pk);
 #if MAX_PK_CACHE_ENTRIES
@@ -414,20 +420,25 @@ get_pubkey_fast (PKT_public_key *pk, u32 *keyid)
       log_error ("keydb_get_keyblock failed: %s\n", g10_errstr(rc));
       return G10ERR_NO_PUBKEY;
     }
-  
+
   assert ( keyblock->pkt->pkttype == PKT_PUBLIC_KEY
            ||  keyblock->pkt->pkttype == PKT_PUBLIC_SUBKEY );
-  copy_public_key (pk, keyblock->pkt->pkt.public_key );
+
+  keyid_from_pk(keyblock->pkt->pkt.public_key,pkid);
+  if(keyid[0]==pkid[0] && keyid[1]==pkid[1])
+    copy_public_key (pk, keyblock->pkt->pkt.public_key );
+  else
+    rc=G10ERR_NO_PUBKEY;
+
   release_kbnode (keyblock);
 
   /* Not caching key here since it won't have all of the fields
      properly set. */
 
-  return 0;
+  return rc;
 }
 
 
-
 KBNODE
 get_pubkeyblock( u32 *keyid )
 {
@@ -1049,7 +1060,12 @@ get_seckey_byname2( GETKEY_CTX *retctx,
 	ctx.items[0].mode = KEYDB_SEARCH_MODE_FIRST;
 	rc = lookup( &ctx, &kb, 1 );
         if (!rc && sk )
-            sk_from_block ( &ctx, sk, kb );
+	  {
+	    sk_from_block ( &ctx, sk, kb );
+	    if(sk->pubkey_algo==PUBKEY_ALGO_ELGAMAL
+	       || (sk->pubkey_algo==PUBKEY_ALGO_ELGAMAL_E && sk->version<4))
+	      rc=G10ERR_UNU_SECKEY;
+	  }
         release_kbnode ( kb );
 	get_seckey_end( &ctx );
     }
@@ -1250,6 +1266,7 @@ fixup_uidnode ( KBNODE uidnode, KBNODE signode, u32 keycreated )
     const byte *p, *sym, *hash, *zip;
     size_t n, nsym, nhash, nzip;
 
+    sig->flags.chosen_selfsig = 1; /* we chose this one */
     uid->created = 0; /* not created == invalid */
     if ( IS_UID_REV ( sig ) ) {
         uid->is_revoked = 1;
@@ -1494,9 +1511,9 @@ merge_selfsigs_main( KBNODE keyblock, int *r_revoked )
     /* pass 1.5: look for key revocation signatures that were not made
        by the key (i.e. did a revocation key issue a revocation for
        us?).  Only bother to do this if there is a revocation key in
-       the first place. */
+       the first place and we're not revoked already. */
 
-    if(pk->revkey)
+    if(!*r_revoked && pk->revkey)
       for(k=keyblock; k && k->pkt->pkttype != PKT_USER_ID; k = k->next )
 	{
 	  if ( k->pkt->pkttype == PKT_SIGNATURE )
@@ -1506,15 +1523,25 @@ merge_selfsigs_main( KBNODE keyblock, int *r_revoked )
 	      if(IS_KEY_REV(sig) &&
 		 (sig->keyid[0]!=kid[0] || sig->keyid[1]!=kid[1]))
 		{ 
-		  /* Failure here means the sig did not verify, is was
+		  int rc=check_revocation_keys(pk,sig);
+		  if(rc==0)
+		    {
+		      *r_revoked=2;
+		      /* don't continue checking since we can't be any
+			 more revoked than this */
+		      break;
+		    }
+		  else if(rc==G10ERR_NO_PUBKEY)
+		    pk->maybe_revoked=1;
+
+		  /* A failure here means the sig did not verify, was
 		     not issued by a revocation key, or a revocation
-		     key loop was broken. */
+		     key loop was broken.  If a revocation key isn't
+		     findable, however, the key might be revoked and
+		     we don't know it. */
 
-		  if(check_revocation_keys(pk,sig)==0)
-		    *r_revoked=1;
-
-		  /* In the future handle subkey and cert revocations?
-                     PGP doesn't, but it's in 2440. */
+		  /* TODO: In the future handle subkey and cert
+                     revocations?  PGP doesn't, but it's in 2440. */
 		}
 	    }
 	}
@@ -1548,19 +1575,21 @@ merge_selfsigs_main( KBNODE keyblock, int *r_revoked )
                      * the same email address may become valid again (hired,
                      * fired, hired again).
                      */
-		    if(sig->flags.expired) {
-		      /* Expired uids don't get to be primary unless
-                         they are the only uid there is. */
-		      uidnode->pkt->pkt.user_id->is_primary=0;
-		      uidnode->pkt->pkt.user_id->is_expired=1;
-		      uidnode->pkt->pkt.user_id->expiredate=sig->expiredate;
-		    }
-                    else {
-                        sigdate = sig->timestamp;
-                        signode = k;
-			if( sig->version > sigversion )
-			  sigversion = sig->version;
-                    }
+		    if(sig->flags.expired)
+		      {
+			uidnode->pkt->pkt.user_id->is_expired=1;
+			signode = NULL;
+		      }
+                    else
+		      {
+			uidnode->pkt->pkt.user_id->is_expired=0;
+			signode = k;
+		      }
+
+		    sigdate = sig->timestamp;
+		    uidnode->pkt->pkt.user_id->expiredate=sig->expiredate;
+		    if( sig->version > sigversion )
+		      sigversion = sig->version;
                 }
             }
         }
@@ -1655,6 +1684,13 @@ merge_selfsigs_main( KBNODE keyblock, int *r_revoked )
         if ( x ) /* mask it down to the actual allowed usage */
             key_usage &= x; 
     }
+
+    /* Type 20 Elgamal keys and the old v3 Elgamal keys are not
+       usable. */
+    if(pk->pubkey_algo==PUBKEY_ALGO_ELGAMAL
+       || (pk->pubkey_algo==PUBKEY_ALGO_ELGAMAL_E && pk->version<4))
+      key_usage=0;
+
     pk->pubkey_usage = key_usage;
 
     if ( !key_expire_seen ) {
@@ -1852,6 +1888,7 @@ merge_selfsigs_subkey( KBNODE keyblock, KBNODE subnode )
 
     subpk->is_valid = 1;
     sig = signode->pkt->pkt.signature;
+    sig->flags.chosen_selfsig=1; /* so we know which selfsig we chose later */
         
     p = parse_sig_subpkt (sig->hashed, SIGSUBPKT_KEY_FLAGS, &n );
     if ( p && n ) {
@@ -1869,6 +1906,14 @@ merge_selfsigs_subkey( KBNODE keyblock, KBNODE subnode )
         if ( x ) /* mask it down to the actual allowed usage */
             key_usage &= x; 
     }
+
+    /* Type 20 Elgamal subkeys, any subkey on a type 20 primary, or
+       any subkey on an old v3 Elgamal(e) primary are not usable. */
+    if(mainpk->pubkey_algo==PUBKEY_ALGO_ELGAMAL
+       || subpk->pubkey_algo==PUBKEY_ALGO_ELGAMAL
+       || (mainpk->pubkey_algo==PUBKEY_ALGO_ELGAMAL_E && mainpk->version<4))
+      key_usage=0;
+
     subpk->pubkey_usage = key_usage;
     
     p = parse_sig_subpkt (sig->hashed, SIGSUBPKT_KEY_EXPIRE, NULL);
@@ -1878,8 +1923,51 @@ merge_selfsigs_subkey( KBNODE keyblock, KBNODE subnode )
         key_expire = 0;
     subpk->has_expired = key_expire >= curtime? 0 : key_expire;
     subpk->expiredate = key_expire;
-}
 
+#ifndef DO_BACKSIGS
+    /* Pretend the backsig is present and accounted for. */
+    subpk->backsig=2;
+#else
+    /* Find the first 0x19 embedded signature on our self-sig. */
+    if(subpk->backsig==0)
+      {
+	int seq=0;
+
+	while((p=enum_sig_subpkt(sig->hashed,
+				 SIGSUBPKT_SIGNATURE,&n,&seq,NULL)))
+	  if(n>3 && ((p[0]==3 && p[2]==0x19) || (p[0]==4 && p[1]==0x19)))
+	    break;
+
+	if(p==NULL)
+	  {
+	    seq=0;
+	    /* It is safe to have this in the unhashed area since the
+	       0x19 is located here for convenience, not security. */
+	    while((p=enum_sig_subpkt(sig->unhashed,SIGSUBPKT_SIGNATURE,
+				     &n,&seq,NULL)))
+	      if(n>3 && ((p[0]==3 && p[2]==0x19) || (p[0]==4 && p[1]==0x19)))
+		break;
+	  }
+
+	if(p)
+	  {
+	    PKT_signature *backsig=m_alloc_clear(sizeof(PKT_signature));
+	    IOBUF backsig_buf=iobuf_temp_with_content(p,n);
+
+	    if(parse_signature(backsig_buf,PKT_SIGNATURE,n,backsig)==0)
+	      {
+		if(check_backsig(mainpk,subpk,backsig)==0)
+		  subpk->backsig=2;
+		else
+		  subpk->backsig=1;
+	      }
+
+	    iobuf_close(backsig_buf);
+	    free_seckey_enc(backsig);
+	  }
+      }
+#endif
+}
 
 
 /* 
@@ -1935,8 +2023,8 @@ merge_selfsigs( KBNODE keyblock )
                 PKT_public_key *pk = k->pkt->pkt.public_key;
 		if(!main_pk->is_valid)
 		  pk->is_valid = 0;
-		if(revoked)
-		  pk->is_revoked = 1;
+		if(revoked && !pk->is_revoked)
+		  pk->is_revoked = revoked;
                 if(main_pk->has_expired)
 		  pk->has_expired = main_pk->has_expired;
             }
@@ -2523,14 +2611,12 @@ get_user_id_string( u32 *keyid )
 
 
 char*
-get_user_id_string_printable ( u32 *keyid )
+get_user_id_string_native ( u32 *keyid )
 {
-    char *p = get_user_id_string( keyid );
-    char *p2 = utf8_to_native( p, strlen(p), 0 );
-    m_free(p);
-    p = make_printable_string (p2, strlen (p2), 0);
-    m_free (p2);
-    return p;
+  char *p = get_user_id_string( keyid );
+  char *p2 = utf8_to_native( p, strlen(p), 0 );
+  m_free(p);
+  return p2;
 }
 
 
@@ -2587,15 +2673,13 @@ get_user_id( u32 *keyid, size_t *rn )
 }
 
 char*
-get_user_id_printable( u32 *keyid )
+get_user_id_native( u32 *keyid )
 {
-    size_t rn;
-    char *p = get_user_id( keyid, &rn );
-    char *p2 = utf8_to_native( p, rn, 0 );
-    m_free(p);
-    p = make_printable_string (p2, strlen (p2), 0);
-    m_free (p2);
-    return p;
+  size_t rn;
+  char *p = get_user_id( keyid, &rn );
+  char *p2 = utf8_to_native( p, rn, 0 );
+  m_free(p);
+  return p2;
 }
 
 KEYDB_HANDLE
